"""Do file io."""
from __future__ import annotations

from os.path import exists, splitext
from pathlib import Path

from ..layeredimage import LayeredImage
from .gif import openLayer_GIF, saveLayer_GIF
from .layered import (
	openLayer_LAYERED,
	openLayer_LAYEREDC,
	saveLayer_LAYERED,
	saveLayer_LAYEREDC,
)
from .lsr import openLayer_LSR, saveLayer_LSR
from .ora import openLayer_ORA, saveLayer_ORA
from .pdn import openLayer_PDN, saveLayer_PDN
from .psd import openLayer_PSD, saveLayer_PSD
from .tiff import openLayer_TIFF, saveLayer_TIFF
from .webp import openLayer_WEBP, saveLayer_WEBP
from .xcf import openLayer_XCF, saveLayer_XCF


def extNotRecognised(fileName: str):
	"""Output the file extension not recognised error."""
	exts = ["ora", "psd", "xcf", "pdn", "tif", "tiff", "webp", "gif", "lsr", "layered", "layeredc"]
	print(
		f"ERROR: File extension is not recognised for file: {fileName}! Must be one of "
		', "'.join(exts) + '"'
	)


def openLayerImage(file: str | Path) -> LayeredImage:
	"""Open a layer image file into a layer image object.

	Args:
		file (str): path/ filename

	Raises:
		FileExistsError: If the layered image does not exist
		ValueError: If the extention is not recognised

	Returns:
		LayeredImage: a layered image object
	"""
	functionMap = {
		"ora": openLayer_ORA,
		"psd": openLayer_PSD,
		"xcf": openLayer_XCF,
		"pdn": openLayer_PDN,
		"tif": openLayer_TIFF,
		"tiff": openLayer_TIFF,
		"webp": openLayer_WEBP,
		"gif": openLayer_GIF,
		"lsr": openLayer_LSR,
		"layered": openLayer_LAYERED,
		"layeredc": openLayer_LAYEREDC,
	}
	if not exists(file):
		print(f"ERROR: {file} does not exist")
		raise FileExistsError
	fileExt = splitext(file)[1].lower()
	if fileExt not in functionMap:
		extNotRecognised(file)
		raise ValueError
	return functionMap[fileExt](file)


def saveLayerImage(fileName: str | Path, layeredImage: LayeredImage) -> None:
	"""Save a layered image to a file.

	Args:
		fileName (str): path/ filename
		layeredImage (LayeredImage): the layered image to save

	Raises:
		ValueError: If the extention is not recognised

	Returns:
		None
	"""
	functionMap = {
		"ora": saveLayer_ORA,
		"psd": saveLayer_PSD,
		"xcf": saveLayer_XCF,
		"pdn": saveLayer_PDN,
		"tif": saveLayer_TIFF,
		"tiff": saveLayer_TIFF,
		"webp": saveLayer_WEBP,
		"gif": saveLayer_GIF,
		"lsr": saveLayer_LSR,
		"layered": saveLayer_LAYERED,
		"layeredc": saveLayer_LAYEREDC,
	}
	fileExt = splitext(file)[1].lower()
	if fileExt not in functionMap:
		extNotRecognised(fileName)
		raise ValueError
	return functionMap[fileExt](fileName, layeredImage)


def exportFlatImage(fileName: str, layeredImage: LayeredImage) -> None:
	"""Export the layered image to a unilayer image file."""
	layeredImage.getFlattenLayers().save(fileName)



"""Do file io - LSR."""
from __future__ import annotations

from os import sep

from blendmodes.imagetools import renderWAlphaOffset

from ..layeredimage import LayeredImage
from ..layergroup import Group, Layer

# pylint: disable=invalid-name
# pylint: disable=import-outside-toplevel


## LSR ##
def openLayer_LSR(file: str) -> LayeredImage:
	"""Open a .lsr file into a layered image."""
	import pylsr

	project = pylsr.read(file)
	groups = []
	for group in project.layers:
		groups.append(
			Group(
				group.name,
				[
					Layer(layer.name, layer.scaledImage(), layer.scaledImage().size)
					for layer in group.images
				],
				group.size,
				(int(group.offsets()[0]), int(group.offsets()[1])),
			)
		)
	return LayeredImage(groups, project.size)


def saveLayer_LSR(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layered image as .lsr."""
	import pylsr

	layers = []
	for group in layeredImage.layersAndGroups:
		if isinstance(group, Layer):
			imageData = [pylsr.LSRImageData(group.image, group.name)]
		else:
			imageData = [
				pylsr.LSRImageData(
					renderWAlphaOffset(layer.image, group.dimensions, layer.opacity, layer.offsets),
					layer.name,
				)
				for layer in group.layers
			]
		layers.append(
			pylsr.LSRLayer(
				imageData,
				group.name,
				group.dimensions,
				(
					group.offsets[0] + group.dimensions[0] // 2,
					group.offsets[1] + group.dimensions[1] // 2,
				),
			)
		)
	lsrImage = pylsr.LSRImage(
		layeredImage.dimensions, fileName.split(sep)[-1].replace(".lsr", ""), layers
	)
	pylsr.write(fileName, lsrImage)



"""Do file io - XCF."""
from __future__ import annotations

from ..blend import BlendType
from ..layeredimage import LayeredImage
from ..layergroup import Group, Layer
from .common import blendModeLookup

# pylint: disable=invalid-name
# pylint: disable=import-outside-toplevel


#### XCF ####
def openLayer_XCF(file: str) -> LayeredImage:
	"""Open an .xcf file into a layered image."""
	from gimpformats.gimpXcfDocument import GimpDocument

	blendLookup = {
		0: BlendType.NORMAL,
		3: BlendType.MULTIPLY,
		4: BlendType.SCREEN,
		5: BlendType.OVERLAY,
		6: BlendType.DIFFERENCE,
		7: BlendType.ADDITIVE,
		8: BlendType.NEGATION,
		9: BlendType.DARKEN,
		10: BlendType.LIGHTEN,
		11: BlendType.HUE,
		12: BlendType.SATURATION,
		13: BlendType.COLOUR,
		14: BlendType.LUMINOSITY,
		15: BlendType.DIVIDE,
		16: BlendType.COLOURDODGE,
		17: BlendType.COLOURBURN,
		18: BlendType.HARDLIGHT,
		19: BlendType.SOFTLIGHT,
		20: BlendType.GRAINEXTRACT,
		21: BlendType.GRAINMERGE,
		23: BlendType.OVERLAY,
		24: BlendType.HUE,
		25: BlendType.SATURATION,
		26: BlendType.COLOUR,
		27: BlendType.LUMINOSITY,
		28: BlendType.NORMAL,
		30: BlendType.MULTIPLY,
		31: BlendType.SCREEN,
		32: BlendType.DIFFERENCE,
		33: BlendType.ADDITIVE,
		34: BlendType.NEGATION,
		35: BlendType.DARKEN,
		36: BlendType.LIGHTEN,
		37: BlendType.HUE,
		38: BlendType.SATURATION,
		39: BlendType.COLOUR,
		40: BlendType.LUMINOSITY,
		41: BlendType.DIVIDE,
		42: BlendType.COLOURDODGE,
		43: BlendType.COLOURBURN,
		44: BlendType.HARDLIGHT,
		45: BlendType.SOFTLIGHT,
		46: BlendType.GRAINEXTRACT,
		47: BlendType.GRAINMERGE,
		48: BlendType.VIVIDLIGHT,
		49: BlendType.PINLIGHT,
		52: BlendType.EXCLUSION,
	}
	project = GimpDocument(file)
	# Iterate the layers and create a list of layers for each group, then remove
	# these from the project layers
	layers = project.layers[::-1]
	index = 0
	groupIndex = 0
	groupLayers = [[]]
	while index < len(layers):
		layerOrGroup = layers[index]
		if layerOrGroup.isGroup:
			index -= 1
			while layers[index].itemPath is not None:
				layer = layers[index]
				groupLayers[groupIndex].append(
					Layer(
						layer.name,
						layer.image,
						(layer.width, layer.height),
						(
							layer.xOffset - layerOrGroup.xOffset,
							layer.yOffset - layerOrGroup.yOffset,
						),
						layer.opacity,
						layer.visible,
						blendModeLookup(layer.blendMode, blendLookup),
					)
				)
				layers.pop(index)
				index -= 1
			index += 2
			groupIndex += 1
			groupLayers.append([])
		else:
			index += 1
	# Iterate the clean project layers and add the group layers in
	groupIndex = 0
	layersAndGroups = []
	for layerOrGroup in layers:
		if layerOrGroup.isGroup:
			layersAndGroups.append(
				Group(
					layerOrGroup.name,
					groupLayers[groupIndex][::-1],
					(layerOrGroup.width, layerOrGroup.height),
					(layerOrGroup.xOffset, layerOrGroup.yOffset),
					layerOrGroup.opacity,
					layerOrGroup.visible,
					blendModeLookup(layerOrGroup.blendMode, blendLookup),
				)
			)
			groupIndex += 1
		else:
			layersAndGroups.append(
				Layer(
					layerOrGroup.name,
					layerOrGroup.image,
					(layerOrGroup.width, layerOrGroup.height),
					(layerOrGroup.xOffset, layerOrGroup.yOffset),
					layerOrGroup.opacity,
					layerOrGroup.visible,
					blendModeLookup(layerOrGroup.blendMode, blendLookup),
				)
			)

	return LayeredImage(layersAndGroups, (project.width, project.height))


def saveLayer_XCF(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layered image as .xcf."""
	del fileName, layeredImage
	print(
		"Saving XCFs is not implemented in gimpformats - "
		"this is a little misleading as functions are present, however these are not "
		"functional"
	)
	raise NotImplementedError



"""Do file io - TIFF."""
from __future__ import annotations

from PIL import Image

from ..layeredimage import LayeredImage
from ..layergroup import Layer
from .common import expandLayersToCanvas

# pylint: disable=invalid-name


#### TIFF ####
def openLayer_TIFF(file: str) -> LayeredImage:
	"""Open a .tiff or a .tif file into a layered image."""
	project = Image.open(file)
	layers = []
	dimensions = [0, 0]
	for index in range(project.n_frames):  # type:ignore
		# Load the correct image
		project.seek(index)
		# Update the project dimensions
		for indx, dimension in enumerate(dimensions):
			if project.size[indx] > dimension:
				dimensions[indx] = project.size[indx]
		ifd = project.ifd.named()  # type:ignore
		# Offsets
		offsetX = 0
		offsetY = 0
		if "XPosition" in ifd:
			offsetX = int(
				ifd["XPosition"][0][0] / ifd["XPosition"][0][1] * ifd["XResolution"][0][0]
			)
		if "YPosition" in ifd:
			offsetY = int(
				ifd["YPosition"][0][0] / ifd["YPosition"][0][1] * ifd["YResolution"][0][0]
			)
		# Add the layer
		layers.append(
			Layer(
				ifd["PageName"][0],
				project.copy(),
				(ifd["ImageWidth"][0], ifd["ImageLength"][0]),
				(offsetX, offsetY),
				1,
				True,
			)
		)
	project.close()
	return LayeredImage(layers, (dimensions[0], dimensions[1]))


def saveLayer_TIFF(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layered image as .tiff or .tif."""
	layers = expandLayersToCanvas(layeredImage, "TIFF")
	layers[0].save(fileName, compression=None, save_all=True, append_images=layers[1:])



"""Do file io - GIF."""
from __future__ import annotations

from PIL import Image

from ..layeredimage import LayeredImage
from ..layergroup import Layer
from .common import expandLayersToCanvas

# pylint: disable=invalid-name


## GIF ##
def openLayer_GIF(file: str) -> LayeredImage:
	"""Open a .gif file into a layered image."""
	project = Image.open(file)
	projectSize = project.size
	layers = []
	for index in range(project.n_frames):
		project.seek(index)
		layers.append(
			Layer(
				f"Frame {len(layers) + 1} ({project.info['duration']}ms)",
				project.copy(),
				projectSize,
			)
		)
	project.close()
	return LayeredImage(layers, projectSize)


def saveLayer_GIF(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layered image as .gif."""
	layers = expandLayersToCanvas(layeredImage, "GIF")
	layers[0].save(
		fileName,
		duration=200,
		append_images=layers[1:],
		version="GIF89a",
		disposal=2,
		save_all=True,
		loop=0,
		transparency=0,
	)



"""Do file io - ORA."""
from __future__ import annotations

from ..blend import BlendType
from ..layeredimage import LayeredImage
from ..layergroup import Group, Layer
from .common import blendModeLookup

# pylint: disable=invalid-name
# pylint: disable=import-outside-toplevel


#### ORA ####
def openLayer_ORA(file: str) -> LayeredImage:
	"""Open an .ora file into a layered image."""
	from pyora import TYPE_LAYER, Project

	blendLookup = {
		"svg:src-over": BlendType.NORMAL,
		"svg:multiply": BlendType.MULTIPLY,
		"svg:color-burn": BlendType.COLOURBURN,
		"svg:color-dodge": BlendType.COLOURDODGE,
		"svg:": BlendType.REFLECT,
		"svg:overlay": BlendType.OVERLAY,
		"svg:difference": BlendType.DIFFERENCE,
		"svg:lighten": BlendType.LIGHTEN,
		"svg:darken": BlendType.DARKEN,
		"svg:screen": BlendType.SCREEN,
		"svg:hard-light": BlendType.HARDLIGHT,
		"svg:soft-light": BlendType.SOFTLIGHT,
		"svg:hue": BlendType.HUE,
		"svg:saturation": BlendType.SATURATION,
		"svg:color": BlendType.COLOUR,
		"svg:luminosity": BlendType.LUMINOSITY,
		"svg:plus": BlendType.ADDITIVE,
		"svg:dst-in": BlendType.DESTIN,
		"svg:dst-out": BlendType.DESTOUT,
		"svg:dst-atop": BlendType.DESTATOP,
		"svg:src-atop": BlendType.SRCATOP,
	}
	layersAndGroups = []
	project = Project.load(file)
	for layerOrGroup in project.children[::-1]:
		if layerOrGroup.type == TYPE_LAYER:
			layersAndGroups.append(
				Layer(
					layerOrGroup.name,
					layerOrGroup.get_image_data(True),
					layerOrGroup.dimensions,
					layerOrGroup.offsets,
					layerOrGroup.opacity,
					layerOrGroup.visible,
					blendModeLookup(layerOrGroup.composite_op, blendLookup),
				)
			)
		else:
			layers = []
			for layer in list(layerOrGroup.children)[::-1]:
				layers.append(
					Layer(
						layer.name,
						layer.get_image_data(True),
						layer.dimensions,
						layer.offsets,
						layer.opacity,
						layer.visible,
						blendModeLookup(layerOrGroup.composite_op, blendLookup),
					)
				)
			layersAndGroups.append(
				Group(
					layerOrGroup.name,
					layers,
					project.dimensions,
					layerOrGroup.offsets,
					layerOrGroup.opacity,
					layerOrGroup.visible,
					blendModeLookup(layerOrGroup.composite_op, blendLookup),
				)
			)
	return LayeredImage(layersAndGroups, project.dimensions)


def saveLayer_ORA(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layered image as .ora."""
	from pyora import Project

	blendLookup = {
		BlendType.NORMAL: "svg:src-over",
		BlendType.MULTIPLY: "svg:multiply",
		BlendType.COLOURBURN: "svg:color-burn",
		BlendType.COLOURDODGE: "svg:color-dodge",
		BlendType.REFLECT: "svg:",
		BlendType.OVERLAY: "svg:overlay",
		BlendType.DIFFERENCE: "svg:difference",
		BlendType.LIGHTEN: "svg:lighten",
		BlendType.DARKEN: "svg:darken",
		BlendType.SCREEN: "svg:screen",
		BlendType.SOFTLIGHT: "svg:soft-light",
		BlendType.HARDLIGHT: "svg:hard-light",
		BlendType.HUE: "svg:hue",
		BlendType.SATURATION: "svg:saturation",
		BlendType.COLOUR: "svg:color",
		BlendType.LUMINOSITY: "svg:luminosity",
		BlendType.ADDITIVE: "svg:plus",
		BlendType.DESTIN: "svg:dst-in",
		BlendType.DESTOUT: "svg:dst-out",
		BlendType.DESTATOP: "svg:dst-atop",
		BlendType.SRCATOP: "svg:src-atop",
	}
	project = Project.new(layeredImage.dimensions[0], layeredImage.dimensions[1])
	for layerOrGroup in layeredImage.layersAndGroups:
		if isinstance(layerOrGroup, Layer):
			project = addLayer_ORA(project, layerOrGroup, blendLookup)
		else:
			group = project.add_group(
				layerOrGroup.name,
				offsets=layerOrGroup.offsets,
				opacity=layerOrGroup.opacity,
				visible=layerOrGroup.visible,
				composite_op=blendModeLookup(layerOrGroup.blendmode, blendLookup, "svg:src-over"),
			)
			for layer in layerOrGroup.layers:
				group = addLayer_ORA(group, layer, blendLookup)
	project.save(fileName)


def addLayer_ORA(project, layer, blendLookup):
	"""Update the project with a shiny new layer."""
	project.add_layer(
		layer.image,
		layer.name,
		offsets=layer.offsets,
		opacity=layer.opacity,
		visible=layer.visible,
		composite_op=blendModeLookup(layer.blendmode, blendLookup, "svg:src-over"),
	)
	return project



"""Do file io - LAYERED(C)."""
from __future__ import annotations

import io
import json
import zipfile
from typing import Any
from zipfile import ZipFile

from PIL import Image

from ..blend import BlendType
from ..layeredimage import LayeredImage
from ..layergroup import Group, Layer
from .common import blendModeLookup

# pylint: disable=invalid-name


## LAYERED ##
def openLayer_LAYERED(file: str) -> LayeredImage:
	"""Open a .layered file into a layered image."""
	blendLookup = {
		"NORMAL": BlendType.NORMAL,
		"MULTIPLY": BlendType.MULTIPLY,
		"ADDITIVE": BlendType.ADDITIVE,
		"COLOURBURN": BlendType.COLOURBURN,
		"COLOURDODGE": BlendType.COLOURDODGE,
		"REFLECT": BlendType.REFLECT,
		"GLOW": BlendType.GLOW,
		"OVERLAY": BlendType.OVERLAY,
		"DIFFERENCE": BlendType.DIFFERENCE,
		"NEGATION": BlendType.NEGATION,
		"LIGHTEN": BlendType.LIGHTEN,
		"DARKEN": BlendType.DARKEN,
		"SCREEN": BlendType.SCREEN,
		"XOR": BlendType.XOR,
		"SOFTLIGHT": BlendType.SOFTLIGHT,
		"HARDLIGHT": BlendType.HARDLIGHT,
		"GRAINEXTRACT": BlendType.GRAINEXTRACT,
		"GRAINMERGE": BlendType.GRAINMERGE,
		"DIVIDE": BlendType.DIVIDE,
		"HUE": BlendType.HUE,
		"SATURATION": BlendType.SATURATION,
		"COLOUR": BlendType.COLOUR,
		"LUMINOSITY": BlendType.LUMINOSITY,
		"PINLIGHT": BlendType.PINLIGHT,
		"VIVIDLIGHT": BlendType.VIVIDLIGHT,
		"EXCLUSION": BlendType.EXCLUSION,
		"DESTIN": BlendType.DESTIN,
		"DESTOUT": BlendType.DESTOUT,
		"DESTATOP": BlendType.DESTATOP,
		"SRCATOP": BlendType.SRCATOP,
	}
	layersAndGroups = []
	with zipfile.ZipFile(file, "r") as layered:
		with layered.open("stack.json") as stackJson:
			stack = json.load(stackJson)
		# Iterate through the layers and groups
		for layerOrGroup in stack["layersAndGroups"]:
			if layerOrGroup["type"] == "LAYER":
				layersAndGroups.append(grabLayer_LAYERED(layered, layerOrGroup, blendLookup))
			else:
				# If its a group grab the layers
				layers = [
					grabLayer_LAYERED(layered, layer, blendLookup)
					for layer in layerOrGroup["layers"]
				]
				layersAndGroups.append(
					Group(
						layerOrGroup["name"],
						layers,
						layerOrGroup["dimensions"],
						layerOrGroup["offsets"],
						layerOrGroup["opacity"],
						layerOrGroup["visible"],
						blendModeLookup(layerOrGroup["blendmode"], blendLookup),
					)
				)
		return LayeredImage(layersAndGroups, stack["dimensions"])


def grabLayer_LAYERED(zipFile: ZipFile, layer: dict[str, Any], blendLookup: dict[str, Any]):
	"""Grab an image from .layered."""
	with zipFile.open("data/" + layer["name"] + ".png") as layerImage:
		image = Image.open(layerImage).convert("RGBA")
	return Layer(
		layer["name"],
		image,
		layer["dimensions"],
		layer["offsets"],
		layer["opacity"],
		layer["visible"],
		blendModeLookup(layer["blendmode"], blendLookup),
	)


def saveLayer_LAYERED(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layered image as .layered."""
	_saveLayer_LAYERED(fileName, layeredImage)


def _saveLayer_LAYERED(fileName: str, layeredImage: LayeredImage, compressed: bool = False) -> None:
	"""Save a layered image as .layered."""
	with zipfile.ZipFile(
		fileName, "w", compression=(zipfile.ZIP_DEFLATED if compressed else zipfile.ZIP_STORED)
	) as layered:
		layered.writestr(
			"stack.json", json.dumps(layeredImage.json(), indent=(None if compressed else True))
		)
		for layer in layeredImage.layers:
			writeImage_LAYERED(layer.image, layered, "data/" + layer.name + ".png", compressed)
		compositeImage = layeredImage.getFlattenLayers()
		thumbnail = compositeImage.copy()
		thumbnail.thumbnail((256, 256))
		imageLookup = {"composite": compositeImage, "thumbnail": thumbnail}
		for imageName, imageData in imageLookup.items():
			writeImage_LAYERED(imageData, layered, imageName + ".png", compressed)


def writeImage_LAYERED(image: Image.Image, zipFile: ZipFile, path: str, compressed: bool = False):
	"""Write an image to the archive."""
	imgByteArr = io.BytesIO()
	imageCopy = image.copy()
	if compressed and len(set(imageCopy.getcolors(maxcolors=256**3))) < 256:
		imageCopy = imageCopy.quantize(colors=256, method=2, kmeans=1)
	imageCopy.save(imgByteArr, format="PNG", optimize=compressed)
	imgByteArr.seek(0)
	zipFile.writestr(path, imgByteArr.read())


## LAYEREDC ##
def openLayer_LAYEREDC(file: str) -> LayeredImage:
	"""Open a .layeredc file into a layered image."""
	return openLayer_LAYERED(file)


def saveLayer_LAYEREDC(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layeredc image as .layered."""
	_saveLayer_LAYERED(fileName, layeredImage, True)



"""Do file io - WEBP."""
from __future__ import annotations

from PIL import Image

from ..layeredimage import LayeredImage
from ..layergroup import Layer
from .common import expandLayersToCanvas

# pylint: disable=invalid-name


## WEBP ##
def openLayer_WEBP(file: str) -> LayeredImage:
	"""Open a .webp file into a layered image."""
	project = Image.open(file)
	projectSize = project.size
	layers = []
	for index in range(project.n_frames):  # type:ignore
		project.seek(index)
		layers.append(Layer(f"Frame {len(layers) + 1}", project.copy(), projectSize))
	project.close()
	return LayeredImage(layers, projectSize)


def saveLayer_WEBP(fileName: str, layeredImage: LayeredImage):
	"""Save a layered image as .webp."""
	layers = expandLayersToCanvas(layeredImage, "WEBP")
	layers[0].save(fileName, duration=200, save_all=True, append_images=layers[1:])



"""Do file io - Common Operations for file readers/writers."""
from __future__ import annotations

from typing import Any

from blendmodes.imagetools import renderWAlphaOffset
from PIL.Image import Image

from ..blend import BlendType
from ..layeredimage import LayeredImage


def blendModeLookup(
	blendmode: Any, blendLookup: dict[Any, Any], default: Any = BlendType.NORMAL
) -> BlendType:
	"""Get the blendmode from a lookup table."""
	if blendmode not in blendLookup:
		print(f"WARNING: {blendmode} is not currently supported!")
		return default
	return blendLookup[blendmode]


def expandLayersToCanvas(layeredImage: LayeredImage, imageFormat: str) -> list[Image]:
	"""Return layers and throw a warning if the image has groups."""
	if len(layeredImage.extractGroups()) > 0:
		print(f"WARNING: {imageFormat}s do not " "support groups so extracting layers")
	layers = []
	for layer in layeredImage.extractLayers():
		layers.append(
			renderWAlphaOffset(layer.image, layeredImage.dimensions, layer.opacity, layer.offsets)
		)
	return layers



"""Base class."""
from __future__ import annotations

from typing import Any

from PIL import Image

from .blend import BlendType

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments


class LayerGroup:
	"""A representation of an image layer or group."""

	def __init__(
		self,
		name: str,
		dimensions: tuple[int, int],
		offsets: tuple[int, int] = (0, 0),
		opacity: float = 1.0,
		visible: bool = True,
		blendmode: BlendType = BlendType.NORMAL,
		**kwargs: Any,
	):
		"""Represent an image layer or group.

		Args:
			name (str): Name of the layer or group
			dimensions ((int, int)): A tuple representing the dimensions in
			pixels
			offsets (tuple, optional): A tuple representing the left and top
			offsets in pixels. Defaults to (0, 0).
			opacity (float, optional): A float representing the alpha value
			where 0 is invisible and 1 is fully visible. Defaults to 1.0.
			visible (bool, optional): Is the layer visible to the user (this
			is often configured per layer or per group by an 'eye' icon).
			Defaults to True.
			blendmode (Blendtype): The blending mode to use. Defaults to BlendType.NORMAL
		"""
		self.name = name
		self.offsets = offsets
		self.opacity = opacity
		self.visible = visible
		self.dimensions = dimensions
		self.blendmode = blendmode
		self.extras = kwargs

	def __repr__(self):
		"""Get the string representation."""
		return self.__str__()

	def __str__(self):
		"""Get the string representation."""
		return (
			"<LayeredImage "
			+ ' "'
			+ self.name
			+ '" ('
			+ str(self.dimensions[0])
			+ "x"
			+ str(self.dimensions[1])
			+ ")>"
		)

	def json(self) -> dict[str, Any]:
		"""Get the object as a dict."""
		return {
			"name": self.name,
			"offsets": self.offsets,
			"opacity": self.opacity,
			"visible": self.visible,
			"dimensions": self.dimensions,
			"blendmode": self.blendmode.name,
		}


class Layer(LayerGroup):
	"""A representation of an image layer."""

	def __init__(
		self,
		name: str,
		image: Image.Image,
		dimensions: tuple[int, int] | None = None,
		offsets: tuple[int, int] = (0, 0),
		opacity: float = 1.0,
		visible: bool = True,
		blendmode: BlendType = BlendType.NORMAL,
	):
		"""Representation of an image layer.

		Args:
			name (str): Name of the layer or group
			image (Image.Image): A PIL Image
			dimensions (tuple[int, int]): A tuple representing the dimensions in
			pixels
			offsets (tuple[int, int], optional): A tuple representing the left and top
			offsets in pixels. Defaults to (0, 0).
			opacity (float, optional): A float representing the alpha value
			where 0 is invisible and 1 is fully visible. Defaults to 1.0.
			visible (bool, optional): Is the layer visible to the user (this
			is often configured per layer or per group by an 'eye' icon).
			Defaults to True.
			blendmode (Blendtype): The blending mode to use. Defaults to BlendType.NORMAL
		"""
		super().__init__(
			name, (0, 0), offsets=offsets, opacity=opacity, visible=visible, blendmode=blendmode
		)
		self.image = image

		# If the user does not specify the dimensions use image.size
		self.dimensions = dimensions or image.size

	def json(self) -> dict[str, Any]:
		"""Get the object as a dict."""
		return {
			"name": self.name,
			"offsets": self.offsets,
			"opacity": self.opacity,
			"visible": self.visible,
			"dimensions": self.dimensions,
			"type": "LAYER",
			"blendmode": self.blendmode.name,
		}


class Group(LayerGroup):
	"""A representation of an image group."""

	def __init__(
		self,
		name: str,
		layers: list[Layer],
		dimensions: tuple[int, int] | None = None,
		offsets: tuple[int, int] = (0, 0),
		opacity: float = 1.0,
		visible: bool = True,
		blendmode: BlendType = BlendType.NORMAL,
	):
		"""Representation of an image group.

		Args:
			name (str): Name of the layer or group
			layers (layeredimage.Layer[]): A list of layers where the next
			index stacks upon the previous layer
			dimensions ((int, int)): A tuple representing the dimensions in
			pixels
			offsets (tuple, optional): A tuple representing the left and top
			offsets in pixels. Defaults to (0, 0).
			opacity (float, optional): A float representing the alpha value
			where 0 is invisible and 1 is fully visible. Defaults to 1.0.
			visible (bool, optional): Is the layer visible to the user (this
			is often configured per layer or per group by an 'eye' icon).
			Defaults to True.
			blendmode (Blendtype): The blending mode to use. Defaults to BlendType.NORMAL
		"""
		# Initialise dimens to 0 and then calculate as below
		super().__init__(
			name, (0, 0), offsets=offsets, opacity=opacity, visible=visible, blendmode=blendmode
		)
		self.layers = layers

		# If the user does not specify the dimensions use the largest x and y of
		# the layers
		self.dimensions = dimensions or (0, 0)
		if dimensions is None:
			layerDimens = [layer.dimensions for layer in layers]
			self.dimensions = (
				max(dimensions[0] for dimensions in layerDimens),
				max(dimensions[1] for dimensions in layerDimens),
			)

	def json(self) -> dict[str, Any]:
		"""Get the object as a dict."""
		layers = [layer.json() for layer in self.layers]
		return {
			"name": self.name,
			"offsets": self.offsets,
			"opacity": self.opacity,
			"visible": self.visible,
			"dimensions": self.dimensions,
			"type": "GROUP",
			"blendmode": self.blendmode.name,
			"layers": layers,
		}



"""LayeredImage class."""
from __future__ import annotations

from typing import Any

# pylint: disable=unused-import
from blendmodes.imagetools import rasterImageOA, rasterImageOffset, renderWAlphaOffset
from deprecation import deprecated
from PIL import Image

from .blend import blendLayers
from .layergroup import Group, Layer


class LayeredImage:
	"""A representation of a layered image such as an ora."""

	def __init__(
		self,
		layersAndGroups: list[Layer | Group],
		dimensions: tuple[int, int] | None = None,
		**kwargs: Any,
	):
		"""LayeredImage - representation of a layered image.

		Args:
			layersAndGroups (list[Layer, Group]): List of layers and groups
			dimensions (tuple[int, int], optional): dimensions of the canvas. Defaults to None.
		"""
		# Write here
		self.layersAndGroups = layersAndGroups
		# Read only
		self.groups = self.extractGroups()
		self.layers = self.extractLayers()
		# If the user does not specify the dimensions use the largest x and y of
		# the layers and groups
		self.dimensions = dimensions or (0, 0)
		if dimensions is None:
			layerDimens = [layerOrGroup.dimensions for layerOrGroup in layersAndGroups]
			self.dimensions = (
				max(layerDimen[0] for layerDimen in layerDimens),
				max(layerDimen[1] for layerDimen in layerDimens),
			)
		self.extras = kwargs

	def __repr__(self):
		"""Get the string representation."""
		return self.__str__()

	def __str__(self):
		"""Get the string representation."""
		return (
			"<LayeredImage ("
			+ str(self.dimensions[0])
			+ "x"
			+ str(self.dimensions[1])
			+ ") with "
			+ str(len(self.layersAndGroups))
			+ " children>"
		)

	def json(self) -> dict[str, Any]:
		"""Get the object as a dict."""
		layersAndGroups = [layerOrGroup.json() for layerOrGroup in self.layersAndGroups]
		return {"dimensions": self.dimensions, "layersAndGroups": layersAndGroups}

	# Get, set and remove layers or groups
	def getLayerOrGroup(self, index: int):
		"""Get a LayerOrGroup."""
		return self.layersAndGroups[index]

	def addLayerOrGroup(self, layerOrGroup: Layer | Group):
		"""Add a LayerOrGroup."""
		self.layersAndGroups.append(layerOrGroup)

	def insertLayerOrGroup(self, layerOrGroup: Layer | Group, index: int):
		"""Insert a LayerOrGroup at a specific index."""
		self.layersAndGroups.insert(index, layerOrGroup)

	def removeLayerOrGroup(self, index: int):
		"""Remove a LayerOrGroup at a specific index."""
		self.layersAndGroups.pop(index)

	# The user may wish to add an image directly
	@deprecated(details="use addImageAsLayer", deprecated_in="2021.2.4")
	def addLayerRaster(
		self, image: Image.Image, name: str
	):  # pylint:disable=missing-function-docstring
		return self.addImageAsLayer(image, name)

	def addImageAsLayer(self, image: Image.Image, name: str):
		"""Resize an image to the canvas and add as a layer."""
		layer = renderWAlphaOffset(image, self.dimensions)
		self.addLayerOrGroup(Layer(name, layer, self.dimensions))

	@deprecated(details="use insertImageAsLayer", deprecated_in="2021.2.4")
	def insertLayerRaster(
		self, image: Image.Image, name: str, index: int
	):  # pylint:disable=missing-function-docstring
		return self.insertImageAsLayer(image, name, index)

	def insertImageAsLayer(self, image: Image.Image, name: str, index: int):
		"""Resize an image to the canvas  and insert the layer."""
		layer = renderWAlphaOffset(image, self.dimensions)
		self.insertLayerOrGroup(Layer(name, layer, self.dimensions), index)

	# The user may want to flatten the layers
	def getFlattenLayers(self, ignoreHidden: bool = True) -> Image.Image:
		"""Return an image for all flattened layers."""
		return flattenAll(self.layersAndGroups, self.dimensions, ignoreHidden)

	def getFlattenTwoLayers(
		self, background: int, foreground: int, ignoreHidden: bool = True
	) -> Image.Image:
		"""Return an image for two flattened layers."""
		flattenedSoFar = flattenLayerOrGroup(
			self.layersAndGroups[background], self.dimensions, ignoreHidden=ignoreHidden
		)
		return flattenLayerOrGroup(
			self.layersAndGroups[foreground],
			self.dimensions,
			flattenedSoFar,
			ignoreHidden=ignoreHidden,
		)

	def flattenTwoLayers(self, background: int, foreground: int, ignoreHidden: bool = True):
		"""Flatten two layers."""
		image = self.getFlattenTwoLayers(background, foreground, ignoreHidden)
		self.removeLayerOrGroup(foreground)
		self.layersAndGroups[background] = Layer(
			self.layersAndGroups[background].name + " (flattened)", image, self.dimensions
		)

	def flattenLayers(self, ignoreHidden: bool = True):
		"""Flatten all layers."""
		image = self.getFlattenLayers(ignoreHidden)
		self.layersAndGroups[0] = Layer(
			self.layersAndGroups[0].name + " (flattened)", image, self.dimensions
		)
		for layer in range(1, len(self.layersAndGroups)):
			self.removeLayerOrGroup(layer)

	# The user may hate groups and just want the layers... or just want the
	# groups
	def extractLayers(self):
		"""Extract the layers from the image."""
		layers = []
		for layerOrGroup in self.layersAndGroups:
			if isinstance(layerOrGroup, Layer):
				layers.append(layerOrGroup)
			else:
				for layer in layerOrGroup.layers:
					# Render the layer
					layers.append(
						Layer(
							layer.name,
							layer.image,
							(
								max(layer.dimensions[0], layerOrGroup.dimensions[0]),
								max(layer.dimensions[1], layerOrGroup.dimensions[1]),
							),
							(
								layerOrGroup.offsets[0] + layer.offsets[0],
								layerOrGroup.offsets[1] + layer.offsets[1],
							),
							layerOrGroup.opacity * layer.opacity,
							layerOrGroup.visible and layer.visible,
						)
					)
		return layers

	def updateLayers(self):
		"""Update the layers from the image."""
		self.layers = self.extractLayers()

	def extractGroups(self):
		"""Extract the groups from the image."""
		return [
			_layerOrGroup
			for _layerOrGroup in self.layersAndGroups
			if isinstance(_layerOrGroup, Group)
		]

	def updateGroups(self):
		"""Update the groups from the image."""
		self.groups = self.extractGroups()


def flattenLayerOrGroup(
	layerOrGroup: Layer | Group,
	imageDimensions: tuple[int, int],
	flattenedSoFar: Image.Image | None = None,
	ignoreHidden: bool = True,
):
	"""Flatten a layer or group on to an image of what has already been flattened.

	Args:
		layerOrGroup (Layer, Group): A layer or a group of layers
		imageDimensions (tuple[int, int]): size of the image
		flattenedSoFar (Image.Image, optional): the image of what has already
		been flattened. Defaults to None.
		ignoreHidden (bool, optional): ignore layers that are hidden. Defaults
		to True.

	Returns:
		Image.Image: Flattened image
	"""
	if ignoreHidden and not layerOrGroup.visible:
		foregroundRender = Image.new("RGBA", imageDimensions)
	elif isinstance(layerOrGroup, Group):
		foregroundRender = renderWAlphaOffset(
			flattenAll(layerOrGroup.layers, imageDimensions, ignoreHidden),
			imageDimensions,
			1,
			layerOrGroup.offsets,
		)
	else:
		# Get a rendered image and apply blending
		foregroundRender = renderWAlphaOffset(
			layerOrGroup.image, imageDimensions, 1, layerOrGroup.offsets
		)
	if flattenedSoFar is None:
		return foregroundRender
	return blendLayers(
		flattenedSoFar, foregroundRender, layerOrGroup.blendmode, layerOrGroup.opacity
	)


def flattenAll(
	layers: list[Layer | Group] | list[Layer],
	imageDimensions: tuple[int, int],
	ignoreHidden: bool = True,
):
	"""Flatten a list of layers and groups.

	Args:
		layers (list[Layer | Group] | list[Layer]): A list of layers and groups
		imageDimensions (tuple[int, int]): size of the image
		been flattened. Defaults to None.
		ignoreHidden (bool, optional): ignore layers that are hidden. Defaults
		to True.

	Returns:
		Image.Image: Flattened image
	"""
	flattenedSoFar = flattenLayerOrGroup(layers[0], imageDimensions, ignoreHidden=ignoreHidden)
	for layer in range(1, len(layers)):
		flattenedSoFar = flattenLayerOrGroup(
			layers[layer], imageDimensions, flattenedSoFar=flattenedSoFar, ignoreHidden=ignoreHidden
		)
	return flattenedSoFar



"""Do file io - PDN."""
from __future__ import annotations

from PIL import Image

from ..blend import BlendType
from ..layeredimage import LayeredImage
from ..layergroup import Layer
from .common import blendModeLookup

# pylint: disable=invalid-name
# pylint: disable=import-outside-toplevel


#### PDN ####
def openLayer_PDN(file: str) -> LayeredImage:
	"""Open a .pdn file into a layered image."""
	from pypdn.reader import BlendType as PDNBlend
	from pypdn.reader import read

	blendLookup = {
		PDNBlend.Normal: BlendType.NORMAL,
		PDNBlend.Multiply: BlendType.MULTIPLY,
		PDNBlend.Additive: BlendType.ADDITIVE,
		PDNBlend.ColorBurn: BlendType.COLOURBURN,
		PDNBlend.ColorDodge: BlendType.COLOURDODGE,
		PDNBlend.Reflect: BlendType.REFLECT,
		PDNBlend.Glow: BlendType.GLOW,
		PDNBlend.Overlay: BlendType.OVERLAY,
		PDNBlend.Difference: BlendType.DIFFERENCE,
		PDNBlend.Negation: BlendType.NEGATION,
		PDNBlend.Lighten: BlendType.LIGHTEN,
		PDNBlend.Darken: BlendType.DARKEN,
		PDNBlend.Screen: BlendType.SCREEN,
		PDNBlend.XOR: BlendType.XOR,
	}
	project = read(file)
	layers = []
	for layer in project.layers:
		image = Image.fromarray(layer.image)
		layers.append(
			Layer(
				layer.name,
				image,
				image.size,
				(0, 0),
				layer.opacity / 255,
				layer.visible,
				blendModeLookup(layer.blendMode, blendLookup),
			)
		)
	return LayeredImage(layers, (project.width, project.height))


def saveLayer_PDN(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layered image as .pdn."""
	del fileName, layeredImage
	print("ERROR: Saving PDNs is not implemented in pypdn")
	raise NotImplementedError



"""Do file io - PSD."""
from __future__ import annotations

from ..blend import BlendType
from ..layeredimage import LayeredImage
from ..layergroup import Group, Layer
from .common import blendModeLookup

# pylint: disable=invalid-name
# pylint: disable=import-outside-toplevel


#### PSD ####
def openLayer_PSD(file: str) -> LayeredImage:
	"""Open a .psd file into a layered image."""
	from psd_tools import PSDImage
	from psd_tools.constants import BlendMode as psdB

	blendLookup = {
		psdB.NORMAL: BlendType.NORMAL,
		psdB.MULTIPLY: BlendType.MULTIPLY,
		psdB.COLOR_BURN: BlendType.COLOURBURN,
		psdB.COLOR_DODGE: BlendType.COLOURDODGE,
		psdB.OVERLAY: BlendType.OVERLAY,
		psdB.DIFFERENCE: BlendType.DIFFERENCE,
		psdB.SUBTRACT: BlendType.NEGATION,
		psdB.LIGHTEN: BlendType.LIGHTEN,
		psdB.DARKEN: BlendType.DARKEN,
		psdB.SCREEN: BlendType.SCREEN,
		psdB.SOFT_LIGHT: BlendType.SOFTLIGHT,
		psdB.HARD_LIGHT: BlendType.HARDLIGHT,
		psdB.EXCLUSION: BlendType.EXCLUSION,
		psdB.HUE: BlendType.HUE,
		psdB.SATURATION: BlendType.SATURATION,
		psdB.COLOR: BlendType.COLOUR,
		psdB.LUMINOSITY: BlendType.LUMINOSITY,
		psdB.DIVIDE: BlendType.DIVIDE,
		psdB.PIN_LIGHT: BlendType.PINLIGHT,
		psdB.VIVID_LIGHT: BlendType.VIVIDLIGHT,
	}
	layersAndGroups = []
	project = PSDImage.open(file)
	for layerOrGroup in project:
		if layerOrGroup.is_group():
			layers = []
			for layer in layerOrGroup:
				layers.append(
					Layer(
						layer.name,
						layer.topil(),
						(layer.width, layer.height),
						(layer.left - layerOrGroup.left, layer.top - layerOrGroup.top),
						layer.opacity / 255,
						layer.visible,
						blendModeLookup(layer.blend_mode, blendLookup),
					)
				)
			layersAndGroups.append(
				Group(
					layerOrGroup.name,
					layers,
					(layerOrGroup.width, layerOrGroup.height),
					(layerOrGroup.left, layerOrGroup.top),
					layerOrGroup.opacity / 255,
					layerOrGroup.visible,
					blendModeLookup(layerOrGroup.blend_mode, blendLookup),
				)
			)
		else:
			layersAndGroups.append(
				Layer(
					layerOrGroup.name,
					layerOrGroup.topil(),
					(layerOrGroup.width, layerOrGroup.height),
					(layerOrGroup.left, layerOrGroup.top),
					layerOrGroup.opacity / 255,
					layerOrGroup.visible,
					blendModeLookup(layerOrGroup.blend_mode, blendLookup),
				)
			)
	return LayeredImage(layersAndGroups, (project.width, project.height))


def saveLayer_PSD(fileName: str, layeredImage: LayeredImage) -> None:
	"""Save a layered image as .psd."""
	del fileName, layeredImage
	print("ERROR: Saving PSDs is not implemented in psd-tools3")
	raise NotImplementedError
